import bpy
bl_info = {
    "name": "face2FACE",
    "author": "Mosh",
    "version": (0, 6, 66),
    "blender": (2, 83, 6),
    "description": "Align two objects by their active faces",
    "warning": "",
    "category": "3D View"}

# Import From Files
if "bpy" in locals():
    print("Reloading face2FACE addon")
    from . import addon
    import importlib
    importlib.reload(addon)
else:
    print("Loading face2FACE addon")
    from . import addon


class F2FPreferences(addon.config.settings.SettingProps, bpy.types.AddonPreferences):
    """ classes consisting of the face2 addon poperties that
        can be altered and stay persistent in the saved file even
        after ending the current session and starting a new session"""
    bl_idname = __package__

    def draw(self, context):

        # edit face2FACE shortcurt
        keymap = context.window_manager.keyconfigs.user.keymaps['3D View']
        keymap_items = keymap.keymap_items
        row = self.layout.row()
        row.label(text='face2FACE shortcut:')
        row.prop(keymap_items['object.face2face'],
                 'type', text='', full_event=True)

        # draw Settings props
        super().draw(context, self.layout, 'Global settings for each new session:')


def register():
    """ register addon and keymaps """
    addon.register()
    bpy.utils.register_class(F2FPreferences)
    addon.config.keymap.create_keymaps(bpy.context)
    print('Registered face2FACE addon')


def unregister():
    """ unregister addon and keymaps """
    addon.unregister()
    bpy.utils.unregister_class(F2FPreferences)
    addon.config.keymap.remove_keymaps(bpy.context)
    print('Unregistered face2FACE addon')
