""" This file contains the operator to activate the face2Face addon"""
import bpy


class OBJECT_OT_DeleteFace2FaceProjection(bpy.types.Operator):
    bl_label = "Delete face2Face projection"
    bl_description = "Delete a face2Face projection by its name (with all objects or data only)"
    bl_idname = "object.delete_face2face_projection"

    projection_x_name: bpy.props.StringProperty(name='projection_x_name',
                                                default="NO_NAME")
    data_only: bpy.props.BoolProperty(name='data_only', default=False)

    target_also: bpy.props.BoolProperty(name='target_also', default=False)

    def execute(self, context):
        # print("deleting face2Face projection:" +
        #      self.projection_x_name+"... -->")
        # if not projection_name is None:
        #    self.projection_x_name = projection_name
        if self.data_only:
            return self.delete_data_only(context)

        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        self.delete_objects_by_projection_name(context, self.projection_x_name)

        self.delete_data_from_target_object_by_projection_name(
            context, self.projection_x_name)
        self.delete_collection_by_projection_name(
            context, self.projection_x_name)
        context.view_layer.update()
        bpy.ops.ed.undo_push(message="face2Face deletion")
        return {'FINISHED'}

    def delete_data_only(self, context):
        self.delete_data_from_target_object_by_projection_name(
            context, self.projection_x_name)
        self.delete_data_from_objects_by_projection_name(
            context, self.projection_x_name)
        return {'FINISHED'}

    def delete_data_from_objects_by_projection_name(self, context, projection_name):
        for obj in context.scene.objects:
            if obj.belongs_to == projection_name:
                obj.belongs_to = "NOTHING"

    def get_collection_by_name(self, col, col_name):
        found = None
        if (col.name == col_name):
            return col
        for col_child in col.children:
            found = self.get_collection_by_name(
                col_child, col_name)
            if found:
                return found

    def delete_collection_by_projection_name(self, context, projection_name):
        main_col = context.scene.collection
        col2del = self.get_collection_by_name(main_col, projection_name)
        if col2del:
            context.scene.collection.children.unlink(col2del)

    def delete_data_from_target_object_by_projection_name(self, context, projection_name):
        for obj in context.scene.objects:
            count = 0
            for projection in obj.projections:
                if projection.projection_name == projection_name:
                    obj.projections.remove(count)
                count += 1

    def get_target_object_by_projection_name(self, context, projection_name):
        for obj in context.scene.objects:
            for projection in obj.projections:
                if projection.projection_name == projection_name:
                    return obj
        return None

    def delete_objects_by_projection_name(self, context, projection_name):
        objects = []
        for obj in context.scene.objects:
            if obj.belongs_to == projection_name:
                objects.append(obj)
        if self.target_also:
            target_obj = self.get_target_object_by_projection_name(
                context, projection_name)
            if not target_obj is None:
                objects.append(target_obj)
        bpy.ops.object.delete({"selected_objects": objects})


def register():
    bpy.utils.register_class(OBJECT_OT_DeleteFace2FaceProjection)


def unregister():
    bpy.utils.unregister_class(OBJECT_OT_DeleteFace2FaceProjection)
