""" This file contains the align faces operator addon"""
import random
import bpy
from bpy.props import EnumProperty
import mathutils
# from bpy_extras.view3d_utils import region_2d_to_vector_3d, region_2d_to_location_3d


# Import From Files
if "bpy" in locals():
    import importlib
    from . import draw
    importlib.reload(draw)
    from . import mesh
    importlib.reload(mesh)
    from . import view
    importlib.reload(view)
    from . import states
    importlib.reload(states)
    from . import ui
    importlib.reload(ui)
    from . ui import hotkey_display
    importlib.reload(hotkey_display)
    from . import helpers
    importlib.reload(helpers)
    from . import config
    importlib.reload(config)
    from . config import settings
    importlib.reload(settings)
    from . config import keymap
    importlib.reload(keymap)
else:
    from . import draw
    from . import mesh
    from . import view
    from . import states
    from . import ui
    from . ui import hotkey_display
    from . import helpers
    from . import config
    from . config import settings
    from . config import keymap


class OBJECT_OT_AlignByFacesModal(bpy.types.Operator):
    """ Align two objects by their active faces """
    bl_label = "face2FACE modal"
    bl_description = "Align two objects interactively by their active faces"
    bl_idname = "object.face2face"
    bl_keymap = (
        ("view3d.select_circle", {"type": 'LEFTMOUSE', "value": 'PRESS'},
         {"properties": [("wait_for_input", False)]}),
        ("view3d.select_circle", {"type": 'LEFTMOUSE', "value": 'PRESS', "ctrl": True},
         {"properties": [("mode", 'SUB'), ("wait_for_input", False)]}),
    )

    execute_mode: bpy.props.StringProperty(
        default="CREATION")
    reactivation_name: bpy.props.StringProperty(
        default="NO_NAME")

    # bl_options = {'REGISTER', 'UNDO'}
    show_gizmo_navigate_start = False
    show_gizmo_tool_start = False
    handler_added = False
    clean = False
    projection_values_stored = False

    status_text = settings.ADDON_NAME+" - IDLE"
    status_text_header_color = settings.colors.STATUS_TEXT_HEADER_COLOR_DEFAULT
    status_text_color = settings.colors.STATUS_TEXT_COLOR_DEFAULT

    # current_state = "IDLE"
    next_state = states.IdleState()
    state = next_state

    target_object = None
    # the different ngon types(tris, quads, etc) of target faces
    target_face_types = set()

    projected_objects = []

    # lists of the projected object sorted by amount of verts
    # all tris and all quads, etc all in one seperate list
    # the index correlates with the amount of verts
    projected_ngons = {}
    projected_types = []

    selected_objects = set()

    # currently active, manipulating object in projected_objects
    object_index = 0

    scale_variation_strength = 0
    scale_relative_strength = 0
    rotation_variation_strength = 0

    target_modes = [("SINGLE", "Single", ""),
                    ("MULTI", "Multi", "")]
    target_mode = "MULTI"

    target_faces = []
    copy_collection = None

    objs_to_move = []
    linked_sources = []

    def determe_target_mode(self, context):
        """ determ if source objects have a single or a multi target """
        self.target_mode = "SINGLE"
        self.target_faces = mesh.get_selected_polygons()

        if len(self.target_faces) > 1:
            self.target_mode = "MULTI"
            # create a collection where to store the copies of the source object
            self.copy_collection = bpy.data.collections.new(settings.current_projection_name)#Source copys")
            context.scene.collection.children.link(self.copy_collection)
        elif context.window_manager.face2face.copy_when_single_target:
            self.copy_collection = bpy.data.collections.new(settings.current_projection_name)#Source copys")
            context.scene.collection.children.link(self.copy_collection)



    """     def copy_obj(self, obj, parent, linked=False, with_children=True):
        copy = obj.copy()
        if not linked:
            copy.data = obj.data.copy()
        if not parent is None:
            copy.parent = parent
            copy.matrix_parent_inverse = obj.matrix_parent_inverse.copy()
        if with_children:
            for child in obj.children:
                self.copy_obj(child, copy, linked)
        return copy

    def link_obj(self, obj, with_children=True):
        self.copy_collection.objects.link(obj)
        if with_children:
            for child in obj.children:
                self.link_obj(child) """


    def align_faces(self, from_obj_orgs, to_obj, context, linked_sources):
        """ align the active face from an object to the active face of a target object """
        # determe the which target face is next ... this is actually a loop but for drawing reasons splitted
        target_face_index = self.target_faces[context.window_manager.f2f_tmp_vars.objects_projected]
        # randomly choose one from the collection of source objects
        object_type = random.randrange(0, len(from_obj_orgs))
        # make a copy from an original source object for each target face
        from_obj_org = from_obj_orgs[object_type]

        if self.target_mode == "MULTI":
            # the projected objects in each group of same objects types have linked meshes sepperate from original source
            if context.window_manager.face2face.linked_copy and not context.window_manager.face2face.linked_from_source:
                # if no blueprint object for the object type exists the create one
                if linked_sources[object_type] is None:
                    from_obj = helpers.copy_obj(context, from_obj_org, None)
                    linked_sources[object_type] = from_obj
                else:  # ...otherwise link from blueprint object
                    from_obj = helpers.copy_obj(context, linked_sources[object_type], None, True)

            else:
                if context.window_manager.face2face.linked_copy:
                    from_obj = helpers.copy_obj(context, from_obj_org, None, True)
                else:
                    from_obj = helpers.copy_obj(context, from_obj_org, None)
            helpers.link_obj(from_obj, self.copy_collection)
        else:
            # for on single object ....
            if context.window_manager.face2face.copy_when_single_target:             
                if context.window_manager.face2face.linked_copy and context.window_manager.face2face.linked_from_source:
                    # ... copy the original source object woth linked data  
                    from_obj = helpers.copy_obj(context, from_obj_org, None, True)        
                else:
                    # ... copy the whole original source object   
                    from_obj = helpers.copy_obj(context, from_obj_org, None)
                helpers.link_obj(from_obj, self.copy_collection)
            else:  
                # ... just take the original source object              
                from_obj = from_obj_org

        #store original matrix
        from_obj.matrix_local_bv = helpers.flatten(
                from_obj.matrix_local)

        from_obj.rotation_mode = 'QUATERNION'

        # orient local transform to target face
        mesh.orient2face(from_obj, to_obj, target_face_index)
        # determe active faces from source and target ... so we can orient the source mesh
        mesh.rotate_mesh2target_face(from_obj, to_obj, target_face_index)

        # Move from_obj so that faces match
        #mesh.position_mesh_face2face.props(from_obj, to_obj, fpoly, tpoly)
        mesh.position_mesh_face2face(
            from_obj, to_obj, target_face_index)
        
        #update children
        helpers.update_children(context, from_obj, from_obj)

        # store target and source face index in source object
        from_obj.target_face_index = target_face_index
        from_obj.source_face_index = mesh.get_selected_polygon(from_obj)
        from_obj.object_type = object_type
        from_obj.select_set(True)

        context.window_manager.f2f_tmp_vars.objects_projected = context.window_manager.f2f_tmp_vars.objects_projected + 1
        context.window_manager.f2f_tmp_vars.options_panel_msg = "PROJECTiNG ... "+str(context.window_manager.f2f_tmp_vars.objects_projected)+"/"+str(len(self.target_faces))
        self.set_status_text(" - " +context.window_manager.f2f_tmp_vars.options_panel_msg)
        bpy.context.area.tag_redraw()

        if context.window_manager.f2f_tmp_vars.objects_projected < len(self.target_faces): 
            return (from_obj, True)
        
        return (from_obj, False) 


    def align_faces_start(self, context):
        #draw.debug_lines = []
        self.determe_target_mode(context)

        # make sure active face is a(the) selected one because blender do not ensure that
        for obj_to_move in self.objs_to_move:
            obj_to_move.data.update()
            mesh.set_active_polygon_to_first_selected(obj_to_move)
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        context.view_layer.update()

        # fill linked  with empty placeholders for each source type object   
        for i in range(0, len(self.objs_to_move)):
            self.linked_sources.append(None)
        

    def set_next_state(self, state_str, blendout_text="", blendout_ret_val='CANCELLED', blendout_color=None):
        if state_str == "MOVE":
            self.next_state = states.GrabState()
        if state_str == "SCALE":
            self.next_state = states.ScaleState()
        if state_str == "ROTATE":
            self.next_state = states.RotateState()
        if state_str == "IDLE":
            self.next_state = states.IdleState()
        if state_str == "BLENDOUT":
            self.next_state = states.BlendOutState()
            self.next_state.ret_val = blendout_ret_val
            self.next_state.text = blendout_text
            if blendout_color:
                self.next_state.text_color = blendout_color

    def switch_state(self, context,  event):
        """ do steps to switch a modal state """
        # leaving_state, entering_state,
        self.state.leave_state(self)
        
        self.state = self.next_state
        hotkey_display.state = self.state

        self.report({'INFO'}, self.state.name)
        self.state.enter_state(context, event, self)

    def set_status_text(self,  text, with_name=True):
        """ set a status text displayed by the draw handler """
        self.status_text = ""
        if with_name:
            self.status_text += settings.ADDON_NAME
        self.status_text += text
        bpy.context.area.tag_redraw()

    def check_requirements(self, context):
        """ check all requirements needed to run this operator  """
        # TODO check if all faces are convex
        # context.object.data.update()
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        bpy.ops.object.mode_set(mode='EDIT', toggle=False)

        if not len(context.selected_objects) >= 2:
            states.BlendOutState.text = "You need at least 2 Selected Objects"
            self.next_state = states.BlendOutState()
            self.next_state.text_color = context.window_manager.face2face.status_text_error_color
            return False

        for obj in context.selected_objects:
            if obj.type != 'MESH':
                states.BlendOutState.text = "All Selected Objects need to be Meshes"
                self.next_state = states.BlendOutState()
                self.next_state.text_color = context.window_manager.face2face.status_text_error_color
                return False

        if len(mesh.get_selected_polygons()) == 0:
            states.BlendOutState.text = "At least 1 target face in the active object has to be selected"
            self.next_state = states.BlendOutState()
            self.next_state.text_color = context.window_manager.face2face.status_text_error_color
            return False

        # determe source objects
        self.objs_to_move = [
            o for o in context.selected_objects if o != context.active_object]

        for obj_to_move in self.objs_to_move:
            obj_to_move.data.update()

            # print("faces in source:",
            #      mesh.count_selected_polygons(objs_to_move[0]))
            if mesh.count_selected_polygons(obj_to_move) != 1:
                states.BlendOutState.text = "excatly 1 face has to be selected in each source object (" + \
                    str(mesh.count_selected_polygons(obj_to_move))+" instead)"
                self.next_state = states.BlendOutState()
                self.next_state.text_color = context.window_manager.face2face.status_text_error_color
                return False

        return True

    def check_requirements_reactivation(self, context):
        """ check all requirements needed to run this operator  """
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        if self.target_object == None:
            states.BlendOutState.text = "The target object do not exists anymore -> projection deleted"
            bpy.ops.object.delete_face2face_projection(data_only=True, projection_x_name=self.reactivation_name)
            self.next_state = states.BlendOutState()
            self.next_state.text_color = context.window_manager.face2face.status_text_error_color
            return False

        if len(self.projected_objects) == 0:
            states.BlendOutState.text = "No projected object exists anymore -> projection deleted"
            bpy.ops.object.delete_face2face_projection(data_only=True, projection_x_name=self.reactivation_name)
            self.next_state = states.BlendOutState()
            self.next_state.text_color = context.window_manager.face2face.status_text_error_color
            return False
        
        #bpy.ops.object.mode_set(mode='EDIT', toggle=False)

        for obj in self.projected_objects:
            #TODO MAKE SURE nothing is hidden unhide it
            # select source face only
            mesh.select_single_polygon(obj, obj.source_face_index)
            # check if source face index exists and decide what to do if not
            try:
                face_index_exists = obj.data.polygons[obj.source_face_index]
            except IndexError:
                states.BlendOutState.text = "The Source face of some objects do not exists anymore"
                self.next_state = states.BlendOutState()
                self.next_state.text_color = context.window_manager.face2face.status_text_error_color
                return False
            # check if target face index exists and decide what to do if not
            try:
                face_index_exists = self.target_object.data.polygons[obj.target_face_index]
            except IndexError:
                states.BlendOutState.text = "The target face of some objects do not exists anymore"
                self.next_state = states.BlendOutState()
                self.next_state.text_color = context.window_manager.face2face.status_text_error_color
                return False
            #TODO give a possibility to assign new ones manually

        return True

    def get_target_face_indicies(self):
        indicies = []
        for obj in self.projected_objects:
            indicies.append(obj.target_face_index)
        return indicies

    def reactivate_target_face_types(self):
        """ set a status text displayed by the draw handler """
        target_face_indicies = self.get_target_face_indicies()
        for target_face_index in target_face_indicies:
            tpoly = mesh.get_polygon(self.target_object, target_face_index)
            tpoly_verts = mesh.get_verts_from_polygon(
                self.target_object, tpoly)
            self.target_face_types.add(len(tpoly_verts))

    def set_target_face_types(self):
        """ set a status text displayed by the draw handler """
        target_faces = mesh.get_selected_polygons()
        for target_face_index in target_faces:
            tpoly = mesh.get_polygon(self.target_object, target_face_index)
            tpoly_verts = mesh.get_verts_from_polygon(
                self.target_object, tpoly)
            self.target_face_types.add(len(tpoly_verts))

    def fill_ngon_lists(self):
        """ fill lists of objects with same target face type """
        for obj in self.projected_objects:
            tpoly = mesh.get_polygon(self.target_object, obj.target_face_index)
            tpoly_verts = mesh.get_verts_from_polygon(
                self.target_object, tpoly)
            index = len(tpoly_verts)
            if not index in self.projected_ngons:
                empty_list = []
                self.projected_ngons.update({index:empty_list})

            self.projected_ngons[index].append(obj)


    def fill_type_lists(self, num_of_types):
        """ fill lists of objects with same target face type """
        for i in range(0, num_of_types):
            # print("ti:", i)
            empty_list = []
            self.projected_types.append(empty_list)

        for obj in self.projected_objects:
            self.projected_types[obj.object_type].append(obj)

    def get_projected_object_by_tft(self, tf_index):
        """ get projected object by its target face index """
        for obj in self.projected_objects:
            if obj.target_face_index == tf_index:
                return obj

    def get_projected_objects_by_tft(self, tf_indicies):
        """ get projected objects by their target face indicies """
        objects = []
        # collect objects by their target face indicies:
        for index in tf_indicies:
            for obj in self.projected_objects:
                if obj.target_face_index == index:
                    objects.append(obj)
        return objects


    def get_current_projection(self):
        """ activates the addon so props are initialized so we can recreate
            former projections  """
        if not self.target_object is None:
            for projection in self.target_object.projections:
                if projection.projection_name == settings.current_projection_name:
                    return projection
        return None


    def store_values_in_current_projection(self):
        """ recreate former projections  """
        projection = self.get_current_projection()
        projection.scale_variation_strength = self.scale_variation_strength
        projection.scale_relative_strength = self.scale_relative_strength
        projection.rotation_variation_strength = self.rotation_variation_strength
        self.projection_values_stored = True

        

    def get_values_from_current_projection(self):
        """ recreate former projections  """
        projection = self.get_current_projection()
        if not projection is None:
            self.get_values_from_projection(projection)

    def get_values_from_projection(self, projection):
        """ recreate former projections  """
        self.scale_variation_strength = projection.scale_variation_strength
        self.scale_relative_strength = projection.scale_relative_strength
        self.rotation_variation_strength = projection.rotation_variation_strength

    def get_target_object_by_projection_name(self, context, projection_name):
        for obj in context.scene.objects:
            for projection in obj.projections:
                if projection.projection_name == projection_name:
                    return obj
        self.report(
            {'WARNING'}, "No target object found for projection named:"+projection_name)
        return None

    def get_projected_objects_by_projection_name(self, context, projection_name):
        objects = []
        for obj in context.scene.objects:
            if obj.belongs_to == projection_name:
                objects.append(obj)
        if len(objects) == 0:
            self.report(
            {'WARNING'}, "No projected objects found for projection named:"+projection_name)     
        return objects

    def reselect_objects(self, context):
        """ select involved objects and deselect all others  """
        for obj in context.scene.objects:
            if obj in self.projected_objects:
                obj.select_set(True)
            else:
                 obj.select_set(False) 

        if not self.target_object is None:
            self.target_object.select_set(False)
        print("pro objs:",len(self.projected_objects))
        if len(self.projected_objects) > 0:
            context.view_layer.objects.active = self.projected_objects[self.object_index]
        #context.view_layer.objects.active = self.target_object
        context.view_layer.update()
    
    def unhide_involved_objects(self, context):
        """ select involved objects and deselect all others  """
        for obj in self.projected_objects:
                obj.hide_set(False)

        if not self.target_object is None:
            self.target_object.hide_set(False)
    
    def adjust_event_timer(self, context, interval):
        if self.handler_added:
            context.window_manager.event_timer_remove(self._timer)
            self._timer = context.window_manager.event_timer_add(
                interval, window=context.window)

    def init_handlers(self, context):
        if not self.handler_added:
            # the arguments we pass to the the callback
            args = (context, self)
            # Add the region OpenGL drawing callback
            self._handle = bpy.types.SpaceView3D.draw_handler_add(
                draw.draw_callback_px, args, 'WINDOW', 'POST_PIXEL')

            self._timer = context.window_manager.event_timer_add(
                0.0025, window=context.window)
            context.window_manager.modal_handler_add(self)

            # undo handler
            args_cu = (self, context)
            if not self.catch_undo in bpy.app.handlers.undo_post:
                bpy.app.handlers.undo_post.append(self.catch_undo)

            self.handler_added = True
            self.set_status_text(" - ")
            bpy.context.area.tag_redraw()

    
    def init(self, context):
        """ init some params and register handler  """
        if context.area.type == 'VIEW_3D':
            self.show_gizmo_navigate_start = context.space_data.show_gizmo_navigate
            context.space_data.show_gizmo_navigate = False
            self.show_gizmo_tool_start = context.space_data.show_gizmo_tool
            context.space_data.show_gizmo_tool = False

            self.linked_sources = []
            self.objs_to_move = []
            self.target_faces = []
            self.copy_collection = None

            self.projected_objects = []
            self.obis = [] #What the obi?
            self.selected_objects = set()
            self.projected_ngons = {}
            self.projected_types = []

            self.target_object = context.active_object
            
            hotkey_display.init()

            # if not already done ... activate addon and init props
            if not context.window_manager.face2face.is_active or True:
                bpy.ops.object.activate_face2face()

            context.window_manager.face2face.right_click_select = keymap.determe_selection_mode(context, self)

            return {'RUNNING_MODAL'}
        self.report({'WARNING'}, "View3D not found, cannot run operator")
        return {'CANCELLED'}



    def init_reactivation(self, context):
            """ init some params and register handler  """
            if context.area.type == 'VIEW_3D':

                self.show_gizmo_navigate_start = context.space_data.show_gizmo_navigate
                context.space_data.show_gizmo_navigate = False
                self.show_gizmo_tool_start = context.space_data.show_gizmo_tool
                context.space_data.show_gizmo_tool = False

                # self.projected_objects = []
                self.selected_objects = set()
                self.projected_ngons = {}
                self.projected_types = []

                #self.set_target_face_types()
                self.reactivate_target_face_types()

                # if not already done ... activate addon and init props
                if not context.window_manager.face2face.is_active or True:
                    bpy.ops.object.activate_face2face()

                context.window_manager.face2face.right_click_select = keymap.determe_selection_mode(context, self)
                # context.window_manager.face2face.determe_selection_mode(context)

                return {'RUNNING_MODAL'}

            self.report({'WARNING'}, "View3D not found, cannot run operator")
            return {'CANCELLED'}

    def reactivate_projection(self, context):

        settings.current_projection_name = self.reactivation_name
        
        # collect involved objects 
        self.target_object = self.get_target_object_by_projection_name(context, self.reactivation_name)
        self.projected_objects = self.get_projected_objects_by_projection_name(context, self.reactivation_name)
        self.unhide_involved_objects(context)
        self.reselect_objects(context)

        self.get_values_from_current_projection()
   
        hotkey_display.init()

        print("pro objs:",len(self.projected_objects))
        if not self.check_requirements_reactivation(context):
            return {'RUNNING_MODAL'}

 
        init_result = self.init_reactivation(context)

        self.selected_objects.update(self.projected_objects)
        self.fill_ngon_lists()
        self.fill_type_lists(len(self.projected_objects))


        self.set_next_state("IDLE")
        self.switch_state(context, None)
        self.state.select_selected(self, context)
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
        self.report({'INFO'}, "reactivated: "+self.reactivation_name+" !!!")
        return init_result

    #hacky tmp array for storing projected objects ... i cant store it directly in self.projected_objects
    #because it destroys performance ... looks like timer event did not behave correctly anymore afterwards
    #could not get the reason at the moment
    obis=[]
    def create_projection(self, context):  
        
        to_obj = context.active_object
        tuple_res = self.align_faces(self.objs_to_move, to_obj, context, self.linked_sources)
        proj_obj = tuple_res[0]
        self.obis.append(proj_obj)
        self.still_creating = tuple_res[1]

        if self.still_creating:   
            return

        self.still_creating = False

        #all projected finish creation
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        from_obj_org = self.objs_to_move[0]# TODO check if this is correct quiet unsure
        context.active_object.select_set(False)
        if self.target_mode == "MULTI" or context.window_manager.face2face.copy_when_single_target:
            for from_obj_org in self.objs_to_move:
                from_obj_org.select_set(False)
     
            
        for proj_obj in self.projected_objects:
                proj_obj.select_set(True)
        context.view_layer.objects.active = from_obj_org

        context.view_layer.update()

        # store projection name in target object
        new_projection = self.target_object.projections.add()
        projection_name = self.target_object.name + \
            ".projection."+str(len(self.target_object.projections))
        new_projection.projection_name = projection_name
        settings.current_projection_name = projection_name
        if self.target_mode == "MULTI" or context.window_manager.face2face.copy_when_single_target:
            self.copy_collection.name = settings.current_projection_name

        # store all projeceted objects in fields for later access
        #self.projected_objects = [o for o in context.selected_objects]
        self.projected_objects += self.obis
        self.selected_objects.update(self.projected_objects)
        self.fill_ngon_lists()
        self.fill_type_lists(len(self.objs_to_move))

        # store original matrix, ct and scaling at start after aligning faces in each obj and set projection_name
        for obj in self.projected_objects:
            obj.belongs_to = projection_name
            obj.matrix_local_startorg = helpers.flatten(
                obj.matrix_local.copy())
             
            #custom_transform = view.get_custom_transform_by_obj(context, obj)
            #obj.matrix_custom_transform = helpers.flatten(
            #    custom_transform.matrix)
            scale_vec = obj.matrix_local.decompose()[2]
            obj.scaling_startorg = scale_vec.copy()
            obj.scaling = scale_vec.copy()
        # set initial variation
        states.ScaleState.set_per_object_variation(self, context)
        states.RotateState.set_per_object_variation(self, context)

        self.set_next_state("IDLE")
        self.switch_state(context, None)


    excecute_init_phase = 0
    started_in_edit_mode = False    
    still_creating = False   
    def execute(self, context):
        """ start function for modal operator  """
        #draw.debug_lines = []
        # avoid starting new projections bevor old one is not cleaned up
        if context.window_manager.face2face.is_modal_active:
            return {'CANCELLED'}
        context.window_manager.face2face.is_modal_active = True
        #print("context.modemode:",context.mode)
        self.started_in_edit_mode = (context.mode == 'EDIT_MESH')
        mesh.face2face_orientation_names = []
        # make sure active is set, current user prefs are scanned and set text color
        self.init_handlers(context)
        context.window_manager.f2f_tmp_vars.objects_projected = 0
        return {'RUNNING_MODAL'}

    def execute_modal(self, context):
        if self.excecute_init_phase == 0:
            self.status_text_header_color = context.window_manager.face2face.status_text_header_color
            self.status_text_color = context.window_manager.face2face.status_text_color
            #self.set_status_text(" - STARTING ...")
            #context.window_manager.f2f_tmp_vars.options_panel_msg = "STARTING ..." 
            self.excecute_init_phase = 1
            return {'RUNNING_MODAL'}
        
        if self.excecute_init_phase < 2:
            self.excecute_init_phase += 1
            return {'RUNNING_MODAL'}

        if self.excecute_init_phase == 2:
            keymap.scan_user_mapping(bpy.context)
            #self.status_text_color = context.window_manager.face2face.status_text_color
            self.excecute_init_phase += 1

        if self.execute_mode == 'CREATION':
            if self.excecute_init_phase == 3:
                ret_val = init_result = self.init(context)
                if 'CANCELLED' in ret_val:
                    end_status = 'ABORTED AT START'
                    self.cleanup(context, end_status)
                    return ret_val
                self.excecute_init_phase += 1
                

            if self.excecute_init_phase == 4:
                if not self.check_requirements(context):
                    self.excecute_init_phase = 6
                    return {'RUNNING_MODAL'}

                self.set_target_face_types()
                self.align_faces_start(context)
                self.excecute_init_phase += 1
                self.still_creating = True

   
            """ while self.still_creating:
                self.create_projection(context, self.objs_to_move) """

            self.create_projection(context)
            if self.still_creating:
                return {'RUNNING_MODAL'}
            
            # make an projected object active
            context.view_layer.objects.active = self.projected_objects[self.object_index]
            ret_val = {'RUNNING_MODAL'}

        elif self.execute_mode == 'REACTIVATION':
            #self.excecute_init_phase += 2
            if self.excecute_init_phase == 3:
                context.window_manager.f2f_tmp_vars.options_panel_msg = "REACTIVATING "+self.reactivation_name+" ..."
                self.set_status_text(" - "+context.window_manager.f2f_tmp_vars.options_panel_msg)
                self.excecute_init_phase += 1
                return {'RUNNING_MODAL'}
            if self.excecute_init_phase == 4:    
                self.excecute_init_phase += 1
                return {'RUNNING_MODAL'}
            ret_val = self.reactivate_projection(context)
        else:
            self.report({'WARNING'}, "execution mode:"+self.execute_mode+" not supported")
            ret_val = {'CANCELLED'}

        if 'CANCELLED' in ret_val:
            end_status = 'ABORTED AT START'
            self.cleanup(context, end_status)

        ret_val = {'RUNNING_MODAL'}

        context.window_manager.f2f_tmp_vars.options_panel_msg = None
        self.adjust_event_timer(context, 0.25)
        self.excecute_init_phase = 6
        return ret_val

    def modal(self, context, event):
        """ handle state switch and delegate event handling to states  """
        # init at start
        if self.excecute_init_phase <= 5:
            return self.execute_modal(context)
        # just avoid invoke event to pass through
        if self.excecute_init_phase ==6:
            self.excecute_init_phase += 1
            return {'RUNNING_MODAL'}
        
        # handle state changes
        if self.state is not self.next_state:
            self.switch_state(context, event)

        # handle the endings ... hopefully they will happy ones
        ret_val = self.state.handle_events(context, event, self)

        # determe the return value of Blendout state and when blendout starts then do cleanup
        if isinstance(self.state, states.BlendOutState):#BlendOutState
            if 'PASS_THROUGH' in ret_val or 'RUNNING_MODAL' in ret_val: #still blending out
                if 'FINISHED' in self.state.ret_val: # will this be finished in the end
                    if not self.projection_values_stored:
                        self.store_values_in_current_projection()
                        
            else: # finally finished
                end_status = 'ABORTED'
                if 'FINISHED' in ret_val:
                    end_status = 'APPLIED'
                if not self.clean: # could already be cleaned by catch undo handler
                     self.cleanup(context, end_status)
                if end_status == 'ABORTED':
                    self.report({'WARNING'}, ""+self.state.text)
                self.report({'INFO'}, ""+self.state.text)

        #  if not Blendout state then just determe the return value of current state
        #  and if modal is done then do cleanup
        elif 'FINISHED' in ret_val or 'CANCELLED' in ret_val:
            print("normal finish cancel, stored:",self.projection_values_stored)
            end_status = 'ABORTED' 
            if 'FINISHED' in ret_val:
                end_status = 'APPLIED'
                if not self.projection_values_stored:
                    self.store_values_in_current_projection()
            if not self.clean:
                self.cleanup(context, end_status)
        return ret_val


    def catch_undo(self, scene, dummy):
        """ user is able to do undo during blendout ... in that case the
            blending show has to stop immediatedly and handlers need to be 
            removed otherwise props could be deleted but still used and 
            crash blender  """
        if isinstance(self.state, states.BlendOutState):

            end_status = 'ABORTED'
            if 'FINISHED' in self.state.ret_val: # will this be finished in the end
                end_status = 'APPLIED'
                if not self.projection_values_stored:
                    self.store_values_in_current_projection()
            self.cleanup(bpy.context, end_status)
            self.state.start_time = 0 # quit immediatedly  
        #mesh.face2face_orientation_names = []
        

    def cleanup(self, context, end_status):
        """ cleanup: remove handlers, reset changed blender and user settings   """
        
        # if end_status == 'ABORTED':
        # TODO else: remove current projection
        # print("abortedself", self)
        """ if hasattr(self, 'target_object'):
            if hasattr(self.target_object, 'projections'):
                for pro in self.target_object.projections:          
                    print("cleanaup aborted pro:",pro.projection_name) """

        self.remove_handlers(context)
        hotkey_display.restore()
        context.space_data.show_gizmo_navigate = self.show_gizmo_navigate_start
        context.space_data.show_gizmo_tool = self.show_gizmo_tool_start
        self.report({'INFO'}, settings.ADDON_NAME+" - "+end_status)
        context.area.tag_redraw()
        context.window_manager.face2face.is_modal_active = False
        self.clean = True
        self.execute_mode = "CREATION"
        
        draw.cleanup(context)
        #print("cleanupp done")
        

    def remove_handlers(self, context):
        """ just remove registered handlers  """
        if self.handler_added:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
            context.window_manager.event_timer_remove(self._timer)
            if self.catch_undo in bpy.app.handlers.undo_post:
                bpy.app.handlers.undo_post.remove(self.catch_undo)

            self.handler_added = False



def register():
    """ register all operator classes """
    bpy.utils.register_class(OBJECT_OT_AlignByFacesModal)


def unregister():
    """ unregister all operator classes """
    bpy.utils.unregister_class(OBJECT_OT_AlignByFacesModal)
