""" collection of custom shaders
    all 3d shaders has a uniform called MVP_Matrix
    which can be set with the most valuable projection matrix
    aka the ModelViewProjection Matrix which transforms a local
    coordinate into screen space """
import gpu

vertex_shader_b = '''
        uniform mat4 MVP_Matrix;

        in vec3 position;
        out vec3 pos;

        void main()
        {
            pos = position;
            gl_Position = MVP_Matrix * vec4(position, 1.0f);
        }
    '''

fragment_shader_b = '''
        uniform float brightness;

        in vec3 pos;

        void main()
        {
            gl_FragColor = vec4(pos * brightness, 1.0);
        }
    '''
rgb_3d_vs = '''
        uniform mat4 MVP_Matrix;

        in vec3 position;

        void main()
        {
            gl_Position = MVP_Matrix * vec4(position, 1.0f);
        }
    '''

rgb_3d_fs = '''
        uniform vec3 color;

        void main()
        {
            gl_FragColor = vec4(color, 1.0);
        }
    '''
rgb_3d = gpu.types.GPUShader(rgb_3d_vs, rgb_3d_fs)

rgb_3dr_vs = '''
        uniform mat4 MVP_Matrix;

        in vec3 position;

        void main()
        {
            gl_Position = MVP_Matrix * vec4(position, 1.0f);
        }
    '''

rgb_3dr_fs = '''
        uniform vec3 color;

        void main()
        {
            gl_FragColor = vec4(color, 1.0);
        }
    '''
rgb_3dr = gpu.types.GPUShader(rgb_3d_vs, rgb_3d_fs)

rgba_3d_vs = '''
        uniform mat4 MVP_Matrix;

        in vec3 position;

        void main()
        {
            gl_Position = MVP_Matrix * vec4(position, 1.0f);
        }
    '''

rgba_3d_fs = '''
        uniform vec4 color;

        void main()
        {
            gl_FragColor = color;
        }
    '''
rgba_3d = gpu.types.GPUShader(rgba_3d_vs, rgba_3d_fs)
