""" is a modal state only for displaying info before leaving enterily """
import time
import bpy

if "bpy" in locals():
    import importlib
    from . import base
    importlib.reload(base)
    from .base import ModalState
    from .. import config
    importlib.reload(config)
    from ..config import settings
    importlib.reload(settings)
    from .. config import keymap
    importlib.reload(keymap)
else:
    # pass
    from . import base
    from .base import ModalState
    from .. import config
    from ..config import settings
    from .. config import keymap


class BlendOutState(ModalState):
    """ is a modal state only for displaying info before leaving enterily """
    name = "BLENDOUT"
    start_time = -1
    secs2show = 5
    secs_min_show = 0.25
    text = ""
    text_color = settings.colors.STATUS_TEXT_HEADER_COLOR_DEFAULT
    ret_val = 'CANCELLED'

    def handle_events(self, context, event, modal):
        """ in the blend out state we just display some text for a while
            before finishing """
        modal.set_status_text(" - "+self.text+" - ")
        modal.status_text_header_color = self.text_color
        if (self.start_time + self.secs2show) < time.time():
            return {self.ret_val}

        if keymap.key_macthes_event(keymap.addon_shortcut_kmi, event):
            self.start_time = 0
            return {'RUNNING_MODAL'}

        if \
                not event.type == "TIMER" and \
                not event.type == "TIMER_REPORT" and \
                not event.type == "MOUSEMOVE" and \
                not event.type == "INBETWEEN_MOUSEMOVE" and \
                not event.type == "NONE":

            self.start_time = 0
            return {'PASS_THROUGH'}

        return {'RUNNING_MODAL'}

    def enter_state(self, context, event, modal):
        """ just set some parameter at the beginning of state """
        context.window_manager.f2f_tmp_vars.options_panel_msg = None
        self.start_time = time.time()
        if self.text_color is settings.colors.STATUS_TEXT_HEADER_COLOR_DEFAULT:
            self.text_color = context.window_manager.face2face.status_text_header_color
            #self.text_h = context.window_manager.face2face.status_text_color
        #draw.blendout = True
        context.window_manager.face2face.blendout = True
        context.window_manager.face2face.do_cleanup = True
        # draw.precleanup()
        # set everything back to starting state
        if modal.started_in_edit_mode:
            bpy.ops.object.mode_set(mode='EDIT', toggle=False)
        else:
            bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

    def leave_state(self, modal):
        """ leaving the state cleanup  """
